from boto import sqs

from twisted.internet import task

from prorab import command, server, config
from prorab.server import events

import threading
import sys
import hashlib
import json


def start():
    for override in sys.argv[1:]:
        k, v = override.split('=')
        default = getattr(config, k)
        if type(default) is int:
            v = int(v)
        elif type(default) is list:
            v = v.split(',')

        setattr(config, k, v)

    conn = sqs.connect_to_region(
        config.region,
        aws_access_key_id=config.id,
        aws_secret_access_key=config.key,
    )

    general_queue = conn.get_queue(config.general_queue)
    personal_queue = conn.get_queue(config.personal_queue)

    def pull_queue():
        if not command.is_working():
            for queue in [personal_queue, general_queue, ]:
                message = queue.read()
                if message:
                    queue.delete_message(message)
                    events.task_pulled(message)
                    command.execute(message.get_body())
                    return

    task.LoopingCall(command.send_out).start(config.SEND_OUT)
    task.LoopingCall(pull_queue).start(config.PULL_QUEUE)
    task.LoopingCall(server.ping).start(config.PING)

    flag = threading.Event()
    threading.Thread(target=command.io_thread, args=(flag, )).start()

    def check_hash(c, m):
        hash = ''.join([a + b for a, b in config.commands.items()])
        if hashlib.md5(hash).hexdigest() != m:
            return {'type': 'error',
                    'message': 'command hash check failed', }
        else:
            return {'type': 'event',
                    'event': 'command-hash-checked', }

    server.register_command('commands_hash', check_hash)

    server.start()
    flag.set()



NAME = 'local1'

# AWS, SQS
region = ''
id = ''
key = ''
queue_prefix = 'dev_prorab_'
general_queue = queue_prefix + 'general'
personal_queue = queue_prefix + NAME


# Auth
ACCESS_TOKEN = 'foobar'


dir = '.'
STDOUT_BUFFER_LENGTH = 32


# Server
WEBSOCKET_PORT = 8099


# Timing things
SEND_OUT = 1
PULL_QUEUE = 2
IO_THREAD = 0.5
PING = 1

try:
    exec(open('/etc/prorab.py').read())
except IOError:
    pass




from prorab.server import send_all


def task_pulled(message):
    pass


@send_all
def task_started(c, m):
    c.sendJson({'type': 'event',
        'event': 'task-started',
        'name': m, })


@send_all
def task_finished(c, p):
    c.sendJson({'type': 'event',
        'event': 'task-finished',
        'code': p, })


@send_all
def task_not_found(c, name):
    c.sendJson({'type': 'error',
        'message': 'task {} notfound'.format(name), })


@send_all
def task_failed(c, exception):
    c.sendJson({'type': 'error',
        'message': 'exception while running task: ' + exception, })



import time
import subprocess
import signal

from prorab.server import events
from prorab import server
from prorab import config

execution_poll = []
stdout_poll = []
pull_poll = []
stdout_buffer = []


def get_process():
    if execution_poll:
        return execution_poll[0]


def pop_process():
    del execution_poll[0]


def is_working():
    return get_process()


def popen(args):
    try:
        p = subprocess.Popen(args, cwd=config.dir, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    except OSError:
        raise RuntimeError('command popen failed!')
    return p


def execute(name, args=None):
    line = config.commands.get(name)
    if not line:
        events.task_not_found(name)
        return

    events.task_started(name)

    for c in server.stdout_subs:
        c.stdout_start = 0

    while True:
        try:
            del pull_poll[0]
            del stdout_buffer[0]
        except IndexError:
            break

    while True:
        try:
            del stdout_buffer[0]
        except IndexError:
            break

    print('Executing {}'.format(line))
    try:
        p = popen(line.split())
    except RuntimeError as e:
        events.task_failed(str(e))
        events.task_finished(1)
        return
    execution_poll.append(p)
    stdout_poll.append(p.stdout)


def send_out():
    for c in server.stdout_subs:
        if c.stdout_start == None:
            c.stdout_start = len(stdout_buffer)

        payload = stdout_buffer[c.stdout_start:]
        if payload:
            c.stdout_start += len(payload)
            c.sendJson({'type': 'stdout', 'payload': ''.join(payload), })


def io_thread(flag):
    while not flag.is_set():
        p = get_process()
        if not p:
            time.sleep(config.IO_THREAD)
            continue

        out = p.stdout
        contents = out.read(config.STDOUT_BUFFER_LENGTH)
        stdout_buffer.append(contents)

        if len(contents) < config.STDOUT_BUFFER_LENGTH:
            stdout_buffer.append('prorab: command finished with {}\n'.format(p.poll()))
            events.task_finished(p.poll())
            pop_process()
            print('Process finished')


def sig_int():
    p = get_process()
    if p:
        print('Interrupted by terminal')
        p.send_signal(signal.SIGINT)
        return {'type': 'event',
                'event': 'task-interrupted', }
    return {'type': 'error',
            'message': 'no active task', }


def sig_kill():
    p = get_process()
    if p:
        print('Terminated by terminal')
        p.terminate()
        return {'type': 'event',
                'event': 'task-terminated', }
    return {'type': 'error',
            'message': 'no active task', }

server.register_command('status', lambda x: {'type': 'status', 'status': 'working' if get_process() else 'idle'})
server.register_command('sigint', lambda x: sig_int())
server.register_command('sigterm', lambda x: sig_kill())



from twisted.internet.protocol import Factory
from prorab.server.helpers import JSONReceiver
from twisted.internet import reactor
from txsockjs.factory import SockJSFactory

import time

from prorab import config


stdout_subs = []
clients = []
commands = {}


def send_all(f):
    def wrap(*args, **kwargs):
        for c in clients:
            f(c, *args, **kwargs)
    return wrap


def call(name, *args, **kwargs):
    if name in commands:
        return commands[name](*args, **kwargs)
    else:
        raise RuntimeError('no such command')


def register_command(name, ifn):
    commands[name] = ifn


class ProrabProtocol(JSONReceiver):
    stdout_start = None

    def connectionMade(self):
        pass

    def connectionLost(self, s):
        try:
            clients.remove(self)
        except ValueError:
            pass
        self.unsubscribe()

    def unsubscribe(self):
        try:
            stdout_subs.remove(self)
            return {'type': 'event',
                    'event': 'unsubscribed', }
        except ValueError:
            pass

    def subscribe(self):
        stdout_subs.append(self)
        return {'type': 'event',
                'event': 'subscribed', }

    def jsonReceived(self, data):
        cmd = data.get('command')
        args = data.get('args', [])

        if cmd == 'auth':
            token = args[0] if len(args) > 0 else None
            if token == config.ACCESS_TOKEN:
                clients.append(self)
                self.sendJson({'type': 'event',
                    'event': 'authorized', })
                return
            else:
                self.sendJson({'type': 'error',
                    'message': 'wrong token', })
                return

        if self not in clients:
            self.sendJson({'type': 'error',
                    'message': 'not authorized', })
        else:
            try:
                self.sendJson(call(cmd, self, *args))
            except RuntimeError as e:
                self.sendJson({'type': 'error',
                        'message': str(e), })


@send_all
def send_task_started(c, m):
    c.sendJson({'type': 'event',
        'event': 'task-started',
        'name': m, })


@send_all
def send_task_finished(c, p):
    c.sendJson({'type': 'event',
        'event': 'task-finished',
        'code': p, })


@send_all
def ping(c):
    c.sendJson({'type': 'ping', })


def start():
    print('Starting server at {}'.format(config.WEBSOCKET_PORT))
    reactor.listenTCP(config.WEBSOCKET_PORT, SockJSFactory(Factory.forProtocol(ProrabProtocol)))
    reactor.run()


register_command('stdout_sub', lambda x: x.subscribe())
register_command('stdout_unsub', lambda x: x.unsubscribe())
register_command('ping', lambda x: {'type': 'ping', 'time': time.time()})



import json

from twisted.protocols.basic import LineReceiver


class JSONReceiver(LineReceiver):
    """
    Class inherited from twisted's LiveReceiver for dealing with json.
    """
    def jsonReceived(self, data):
        """
        Called when lineReceived, but deals with json instead of text.

        data -- json value.
        """
        pass

    def sendJson(self, data):
        """
        Send json. Dump data and sendLine.

        data -- value to dump.
        """
        self.sendLine(json.dumps(data).encode("utf-8"))

    def lineReceived(self, line):
        self.jsonReceived(json.loads(line.decode("utf-8")))



#!/usr/bin/env python

from distutils.core import setup

setup(name='prorab',
      version='0.1.0.4',
      description='no description',
      author='Vasiliy Horbachenko',
      author_email='shadow.prince@ya.ru',
      url='',
      scripts=['prorab/bin/prorab', 'prorab/bin/prorabd', ],
      packages=['prorab', 'prorab.server', ],
     )



from boto import sqs
from boto.sqs.message import Message

from prorab import config

conn = sqs.connect_to_region(
    config.region,
    aws_access_key_id=config.id,
    aws_secret_access_key=config.key,
)


def personal(queue, message, args=''):
    q = conn.get_queue(config.queue_prefix + queue)

    m = Message()
    m.set_body(message)
    q.write(m)


def general(message):
    q = conn.get_queue(config.general_queue)

    m = Message()
    m.set_body(message)
    q.write(m)
